/* vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv */
/*  Copyright (c) 1988; Jon R. Herbold and Richard R. A. Shoobridge         */
/*  Indianapolis, IN 46240 ALL RIGHTS RESERVED                              */
/*                                                                          */
/*  FILE NAME:                  FUNCTION NAME:                              */
/*     basecvt.c                   int = nBaseConvert( )                    */
/*                                                                          */
/*  DESCRIPTION:                                                            */
/*     This function converts numbers from one base to another base.  Both  */
/*     bases must be in the range of 2 - 36.  Additional parameters allow   */
/*     for the formating of the output.                                     */
/*                                                                          */
/*  PARAMETERS:                                                             */
/*    char    *cpNumberToConvert;      ASCIIZ number to be converted        */
/*    int     nSourceBase;             base of cpNumberToConvert            */
/*    int     nDestinationBase;        result base                          */
/*    char    *cpResult;               ASCIIZ result of conversion.  Must   */
/*                                     be MAX_LENGTH_OUTPUT+1 chars long.   */
/*    int     nSeparateOutput;         This parameters specifies whether or */
/*                                     not the output is to be separated.   */
/*                                     For example converting 65535 to base */
/*                                     16 would be FFFF if the output is not*/
/*                                     to be separated; otherwise it would  */
/*                                     output FF FF.                        */
/*    int     cSeparationCharacter;    If nSeparateOutput was defined as    */
/*                                     TRUE, then this is the character that*/
/*                                     would separate the groups.  For      */
/*                                     example, if cSeparationCharacter were*/
/*                                     "-", then the output would be FF-FF. */
/*    int     nPadToLeft;              This parameter determines whether or */
/*                                     not the output should be padded to   */
/*                                     the left with zero's.  For example,  */
/*                                     F FF would become 0F FF.             */
/*    int     *npError;                   returns any errors                */
/*                                                                          */
/* Compilers: Microsoft 5.1                                                 */
/*            TurboC 2.0                                                    */
/*                                                                          */
/* Memory Model: any                                                        */
/*                                                                          */
/*                                                                          */
/* Switches: TEST - if == 1, a test driver is compiled                      */
/*           USE_FLOAT - if == 1, then conversion achieves greater          */
/*                       range by using floating point to hold the          */
/*                       intermediate results                               */
/*                                                                          */
/* ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ */

#define     TEST            1   /* compile in test driver */
#define     USE_FLOAT       0   /* use floating point */ 


#if USE_FLOAT == 1
#define TEMP long double
#else
#define TEMP unsigned long
#endif

#include    <stdio.h>
#include    <stdlib.h>
#include    <string.h>
#include    <ctype.h>

#define     BYTE                        unsigned char
#define     WORD                        unsigned int
#define     NO_ERROR                    0
#define     ERROR                       !NO_ERROR
#define     INVALID_SOURCE_BASE         1
#define     INVALID_DESTINATION_BASE    2
#define     INVALID_CHAR_FOR_BASE       3
#define     INVALID_TOO_LARGE           4

#define     MAX_LENGTH_OUTPUT           80

/* RANGE OF BASES THE ROUTINE HANDLES */
#define     LOWER_LIMIT                 2
#define     UPPER_LIMIT                 36

int nBaseConvert(
    char    *cpNumberToConvert,
    int     nSourceBase,
    int     nDestinationBase,
    char    *cpResult,
    int     nSeparateOutput,
    int     cSeparationCharacter,
    int     nPadToLeft,
    int     *npError)
{
    TEMP    TempValue1;
    TEMP    TempWidth;
    WORD    wTempValue2;
    WORD    wTemp;
    WORD    wLoop;
    char    caTempArray[2*MAX_LENGTH_OUTPUT + 1];
    char    *holding;
    int     nReturnValue;

#if USE_FLOAT == 1
    WORD    wCounter;
    WORD    wNumberOfPlaces;
#endif

    /* Array that defines how we group output.  Ex: bBatchArray[2] == 8,
       so we output base 2 stuff in blocks of 8 digits */
    static BYTE    bBatchArray[] =
                   { 0, 0, 8, 0, 0, 0, 0, 0, 3, 0,
                     3, 0, 0, 0, 0, 0, 2, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                     0, 0, 0, 0, 0, 0, 0 };

    /* INITIALIZE THE ERROR FLAGS AND RETURN VALUES */
    *npError = NO_ERROR;
    nReturnValue = NO_ERROR;
    *cpResult = '\0';

    /* BASES MUST BE BETWEEN LOWER_LIMIT AND UPPER_LIMIT */
    if( (nSourceBase < LOWER_LIMIT) || (nSourceBase > UPPER_LIMIT) )
    {
        *npError  = INVALID_SOURCE_BASE;
        nReturnValue = ERROR;
    }
    else if( (nDestinationBase < LOWER_LIMIT) || (nDestinationBase > UPPER_LIMIT) )
    {
        *npError  = INVALID_DESTINATION_BASE;
        nReturnValue = ERROR;
    }
    else
    {
        /* REVERSE ORDER - WE WORK FROM LEAST TO MOST SIGNIFICANT DIGIT */
        strrev( cpNumberToConvert );

        /* PUT BINARY cpNumberToConvert INTO wTempValue1 */
        for( wLoop=0, TempValue1=(TEMP) 0, TempWidth=(TEMP) 1;
            cpNumberToConvert[wLoop];
            wLoop++ )
        {

            /* CONVERT CHARACTER TO NUMERIC EQUIVALENT */
            wTempValue2 = UPPER_LIMIT;
            if (isdigit(cpNumberToConvert[wLoop]))
                wTempValue2 = cpNumberToConvert[wLoop] - '0';
            else if (isalpha(cpNumberToConvert[wLoop]))
                /* 'A' - '7' gives 10 ... */
                wTempValue2 = toupper(cpNumberToConvert[wLoop]) - '7';

            if( wTempValue2 >= nSourceBase )
            {
                *npError  = INVALID_CHAR_FOR_BASE;
                nReturnValue = ERROR;
                break;
            }
            else
            {
                /* UPDATE TEMPORARY VALUE AND WIDTH */
                TempValue1 += ((TEMP) wTempValue2) * TempWidth;
                TempWidth *= (TEMP) nSourceBase;
            }
        } /* end of for loop */

        /* RESTORE TO ORIGINAL ORDER */
        strrev( cpNumberToConvert );

        /* CHECK FOR POSSIBLE CONVERSION ERROR */
        if( *npError == NO_ERROR )
        {

#if USE_FLOAT == 1
            /* FIND HIGHEST DIVISOR */
            for( TempWidth=(TEMP) 1, wNumberOfPlaces=1;
                (TempWidth * (TEMP)nDestinationBase) <= TempValue1; )
            {
                TempWidth *= (TEMP)nDestinationBase;
                wNumberOfPlaces++;
            }

            /* CONVERT TempValue1 TO DESTINATION BASE & PUT INTO caTempArray */
            for( wLoop=0,wCounter=0; wNumberOfPlaces; wLoop++, wCounter++, wNumberOfPlaces-- )
            {
                wTemp = (WORD) (TempValue1 / TempWidth);
                caTempArray[wLoop] = (char)(wTemp<10 ? wTemp+'0' : wTemp+'7' );
                TempValue1 = TempValue1 - ((TEMP) wTemp * TempWidth );
                TempWidth /= (TEMP) nDestinationBase;
                if( wLoop >= MAX_LENGTH_OUTPUT - 1 )
                {
                    *npError  = INVALID_TOO_LARGE;
                    *caTempArray = '\0';
                    return( ERROR );
                }
            }
            /* NULL TERMINATE AND REVERSE THE RESULT STRING */
            caTempArray[wLoop] = '\0';
            strrev( caTempArray );

#else
            /* CONVERT TempValue1 TO DESTINATION BASE & PUT INTO caTempArray */
            wLoop=0;
            while(TempValue1 != 0)
            {
                wTemp = (WORD) (TempValue1 % (TEMP) nDestinationBase);
                TempValue1 /= (TEMP) nDestinationBase;
                caTempArray[wLoop++] = (char)(wTemp<10 ?
                    wTemp+'0' : wTemp+'7' );
                if( wLoop >= MAX_LENGTH_OUTPUT - 1 )
                {
                    *npError  = INVALID_TOO_LARGE;
                    return( ERROR );
                }
            }
            /* NULL TERMINATE THE RESULT STRING */
            caTempArray[wLoop] = '\0';
#endif

            /* DO ANY PADDING IF SPECIFIED */
            if( nPadToLeft && bBatchArray[nDestinationBase] )
            {
                wLoop = (WORD) strlen( caTempArray );
                while( (wLoop % bBatchArray[nDestinationBase]) )
                    caTempArray[wLoop++] = '0';
                caTempArray[wLoop] = '\0';
            }

            /* NOW GROUP VALUES IN "caTempArray" IF SPECIFIED */
            if( nSeparateOutput && bBatchArray[nDestinationBase] )
            {
                holding = caTempArray;
                for( wLoop=1; *holding; holding++, wLoop++ )
                {
                    if( !(wLoop % bBatchArray[nDestinationBase]) )
                    {
                        memmove( holding+2, holding+1, strlen( holding+1 ) + 1 );
                        *++holding = (char) cSeparationCharacter;
                        wLoop = 0;
                    }
                }
                if( *--holding == (char) cSeparationCharacter )
                    *holding = '\0';
            }
            if( strlen(caTempArray) >= MAX_LENGTH_OUTPUT )
            {
                *npError  = INVALID_TOO_LARGE;
                nReturnValue = ERROR;
            }
            else
            {
                strcpy( cpResult, caTempArray );
                strrev( cpResult );
            }

        } /* END OF IF NO_ERROR */
    } /* END OF ELSE NO_ERROR */

    return( nReturnValue );
}


#if TEST==1
int main( int argc, char **argv)
{
    char    result[80]; /* holds result of conversion       */
    int     error_code; /* holds error code         */
    char    *cpPointer;

    if( argc != 6 && argc != 7 )
    {
        printf( "BASECVT - convert a number from one base to another for bases 2 - 36\n" );
        printf(" FORMAT:\nbasecvt number src_base dest_base separate fill_out sep_char\n" );
        printf("         number      - the number to be converted\n");
        printf("         src_base    - the current base of 'number'\n");
        printf("         dest_base   - the desired destination base\n");
        printf("         separate    - 1 = print result in groupings (e.g. FF FF)\n");
        printf("                       else 0 = do not group (e.g., FFFF)\n");
        printf("         fill_out    - 1=pad with leading '0', else 0=no pad\n");
        printf("         sep_char    - (optional): charcter that separates groups of digits.\n");
        printf("                       The default is ' '.\n");
        return( 1 );
    }
    nBaseConvert( argv[1], atoi(argv[2]), atoi(argv[3]), result,
        atoi(argv[4]), 
        argc == 6 ? ' ' : *argv[6],
        atoi(argv[5]), &error_code );

    cpPointer = NULL;

    switch( error_code )
    {
        case    INVALID_SOURCE_BASE:
            cpPointer = "Invalid source base";
            break;
        case    INVALID_DESTINATION_BASE:
            cpPointer = "Invalid destination base";
            break;
        case    INVALID_CHAR_FOR_BASE:
            cpPointer = "Invalid character for source base";
            break;
        case    INVALID_TOO_LARGE:
            cpPointer = "Invalid result too large";
            break;
    }

    if( cpPointer == NULL )
        printf( "Result = %s\n", result );
    else
        printf( "%s\n", cpPointer );
}
#endif
