/********************************* BGIPARSE.C ***************************
 * Author: John Dlugosz.  Adapted to Turbo C by John Rex
 * Compiler: Turbo C 2.0
 *
 * Purpose: parses contents of an environment variable to
 *          select various video parameters.  The variable should
 *          contain these elements:
 *
 *          <path>[;<name or number>[;<mode>]]
 *
 *  where:  path is a directory spec where the .bgi/.chr files are
 *          name/number is the name/number of the desired .bgi driver
 *          mode is the desired video display mode.
 *
 * Syntax:
 *
 *  void bgiparse(char *var, char *path, int *drivernum, int *mode)
 *
 *     If a mode is present, it is returned in mode, else mode is set
 *       to zero.
 *     If a driver number is present, it is placed in drivernum.
 *     If a driver name is present, drivernum is set to -1 and the name
 *       of the driver is returned as a part of path (i.e., the whole
 *       name is formed.)
 *     The path is placed in path.
 *
 * Compile time switches:
 *     DEBUG compiles in a test driver
 *
 * Implementation note:  This code turned up a problem with the strtok()
 *  function.  It fails when parsing ";;;" using a ";" for the delimeter.
 *  This is not a bug in the Turbo C implementation, but rather a result
 *  of how the function is defined by ANSI.  Each use of strtok() must step
 *  around this problem.
 *************************************************************************/

#include <stdlib.h>
#include <ctype.h>
#include <string.h>

void bgiparse(char *var, char *path, int *drivernum, int *mode)
{
    char varcopy[200];      /* to hold a copy of var */
    char *delim = ";";       /* character to delimit var */
    char *ptr;

    /* set up default responses */
    *path='\0';
    *drivernum = 0 ;    /* autodetect */
    *mode = 0;
    if (var == NULL) return;

    /* make a copy of var that strtok() can tear apart */
    strcpy(varcopy, var);

    /* first, the path */
    if (*varcopy == *delim)
        { *varcopy = '\0'; ptr = varcopy; }
    else
        ptr = strtok(varcopy, delim);

    if (ptr == NULL) return;
    strcpy(path, ptr);

    /* next the driver name or number */
    ptr = ptr + strlen(ptr) + 1;    /* advance to next potential token */
    if (*ptr == *delim)
        { *ptr = '\0'; ptr = ptr; }
    else
        ptr = strtok(ptr, delim);

    if (ptr == NULL) return;
    if (isdigit(*ptr))
        *drivernum = atoi(ptr);
    else if (*ptr != '\0') {  /* must be a name */
        /* ensure trailing \ */
        if (*(path + strlen(path) - 1) != '\\')
            strcat(path, "\\");
        strcat(path, ptr);
        *drivernum = -1;    /* report that we have a full name */
    }

    /* lastly, a mode */
    ptr = ptr + strlen(ptr) + 1;    /* advance to next potential token */
    if (*ptr == *delim)
        { *ptr = '\0'; ptr = ptr; }
    else
        ptr = strtok(ptr, delim);

    if (ptr == NULL) return;
    *mode = atoi(ptr);

    return;
}

#if defined (DEBUG)

main()
{
    char path[100];
    int drivernum;
    int mode;

    bgiparse(getenv("BGI"), path, &drivernum, &mode);

    printf("BGI contained %s.  It parsed to:\n", getenv("BGI"));
    printf("path = %s, drivernum = %d, mode = %d\n", path, drivernum, mode);
}
#endif
