/***********************     MMATCH. C     ***********************
 * Author: John Rex
 * Compilers: Turbo C 2.0, MSC 5.1, DeSmet 3.0, MetaWare 1.4, Watcom 7
 * Usage:  mmatch.c implements a MUMPS-like matching operator which is
 *         fully explained in the accompanying article.  Its syntax is
 *
 *        int mmatch(char *, char*);
 *
 *        ret=mmatch(pattern,string);
 *
 *  On return, ret will be 1 (True) if string matched pattern,
 *  otherwise it will be 0 (False).  The function returns -1
 *  if the pattern is invalid.  For example,
 *
 *        ret=mmatch("1n.n1'-'1n.n1'-'1n.n",string);
 *
 *  would test to see if string is of the form nn-nn-nn, or a date.
 *
 *  If DRIVER is defined or equal to 1, a test driver is included
 *
 * Source code may be freely used if authorship is acknowledged.
 * Object code form may be freely used. (c) 1989 The C Gazette
 ****************************************************************/

#define DRIVER 1

#if !defined(DRIVER)
  #define DRIVER 0
#endif

#define QUOTE '\''

#include <ctype.h>
#include <stdlib.h>

int mmatch(char *pattern, char *string)
{
    unsigned minrep, maxrep, count;
    int ok;
    char *p_anchor, *s_anchor;

    ok = 1;
    while (*pattern != '\0')
    {

        /* first, get a rep counts:
           must allow for these forms:  .   => 0 to infinity
                                        1   => exactly one
                                        1.3 => one to three
                                        .3  => 0 to three
        */

        minrep = (unsigned) atoi(pattern);  /* pull off first number */
        maxrep = minrep;
        while (isdigit(*pattern))       /* and skip past it */
            pattern++;
        if (*pattern == '.')
        {          /* . operator? */
            pattern++;
            maxrep=65535;

            if (isdigit(*pattern))
            {                         /* trailing count? */
                maxrep = (unsigned) atoi(pattern);
                while (isdigit(*pattern))
                    pattern++;
            }
        }

        /* now start matching */
        p_anchor = pattern;           /* remember where to return to */

        count = 0;
        for(;;)
        {
            pattern = p_anchor;       /* wind up and start testing */
            s_anchor = string;

            if (*pattern == QUOTE)
            {    /* a string to match */
                ok = 1;
                pattern++;
                while (*pattern != QUOTE && *pattern != '\0')
                {
                    if (*pattern == '\\') /* account for escaped chars */
                        pattern++;

                    if (!ok)
                        ;
                    else if (*pattern != *string || *string == '\0')
                        ok = 0; /* say we failed */
                    else
                        string++;

                    pattern++;
                }
                if (*pattern == QUOTE)
                    pattern++;          /* eat the trailing quote */
                else
                    ok = -1;
            }
            else
            { /* type code(s) to match */
                ok = 0;
                while (isalpha(*pattern) && *pattern != '\0')
                {
                    /* try multiple possiblities */
                    if (!ok) switch (toupper (*pattern))
                    {
                        case 'N':
                        case 'D':
                            if (isdigit(*string))
                                ok = 1;
                            break;

                        case 'A':
                            if (isalpha(*string))
                                ok = 1;
                            break;

                        case 'L':
                            if (islower(*string))
                                ok = 1;
                            break;

                        case 'U':
                            if (isupper(*string))
                                ok = 1;
                            break;

                        case 'P':
                            if (ispunct(*string))
                                ok = 1;
                            break;

                        case 'E':
                            if (*string != '\0')
                                ok = 1;
                            break;

                        default:
                            ok = -1; /* bad code */
                            break;
                    }
                    pattern++;
                }
                if (ok == 1)
                    string++;
            }

            /* We've now made a pass at matching the string */
            /* How did we do? */

            if (ok == -1)
                return (-1);        /* bad code */

            else if (ok == 1)
            {     /* found a match */
                count++;            /* count it */
                if (count == maxrep)
                    break;          /* we're done */
            }
            else
            {                  /* ok == 0 -- no match */
                /* are we within range? */
                if (minrep <= count && count <= maxrep)
                { /* yes */
                    string = s_anchor;
                    ok = 1;         /* it's really ok */
                }
                break;
            }
        }

        /* we just finished testing <count><pattern> */
        /* how did we do? */
        if (!ok)
            break;
    }

    /* if we are at the end of string, we succeeded! */
    if (*string == '\0' && *pattern == '\0' && ok)
        return (1);
    else
        return (0);
}

#pragma page()     /* Portability: comment out this pragma for MetaWare */

#if defined(DRIVER)

#include <stdio.h>

void main(int argc, char **argv)
{
    int ret;

    if (argc == 1 || argc > 3)
    {
        puts("Usage is: mmatch pattern [string]\n");
        puts("  If string is omitted, a "" string is used");
        return;
    }

    if (argc == 3)
        ret = mmatch(argv[1], argv[2]);
    else
        ret = mmatch(argv[1], "");

    printf("mmatch() returned %d\n", ret);
}
#endif