

/* Serial support for target communications. Assumes 8250 serial device */

#include <dos.h>
#include <stdio.h>
#include <string.h>
#include <i32.h>
#include <process.h>
#include "ttycntl.h"
#include "gnudos.h"

static char serial_device[20];
static int  serial_baud;
static int  serial_devtype; /* Device identifier. */
static int  serial_iobase;  /* I/O base value. */
static int  serial_ivec;    /* Host interrupt vector. */
static char serial_opened;  /* Flag set if file is opened */
static char serial_setup;   /* Flag set if setup for port has started */
static char serial_updated; /* Flag set if port values were altered */

static struct dupdef {
	int num;
	int max;
	int *fd;
} serial_dup = { 0, 5, NULL };

static void (*orig_serial_vec)() = 0;

#define IR3		0x0b
#define IR4		0x0c
#define IR5		0x0d

#define IMR_8259	0x21		/* int mask reg */

#define LOOPERMS	33		/* rough loop count per ms */
#define BREAKMS 	1000		/* break length in ms */

/*
 * Map bauds onto counter values
 */
static struct baudmap {
	unsigned long baud;
	int code;
} baudmap[] ={
	{   1200, 0x060},
	{   2400, 0x030},
	{   4800, 0x018},
	{   9600, 0x00C},
	{  19200, 6},
	{  38400, 3},
	{  57600, 2},
	{ 115200, 1},
	{ 166600, -1},
	{ 250000, -1},
	{ 500000, -1},
	{ 750000, -1},
	{1500000, -1},
	{      0, 0}
};

/*
 * Input queue
 */
#define QBUFSZ	(3*1024)	/* queue size */

static struct queue {
	int		q_cnt;		/* amount in buffer	*/
	unsigned	q_ovfl; 	/* number of overflows	*/
	char *		q_in;		/* next free spot	*/
	char *		q_out;		/* next char out	*/
	char		q_buf[QBUFSZ];	/* character buffer	*/
} siq;		/* chars from remote port */


static int  get_port(int comwhich);
static void init_8259(unsigned int int_value);
static int  is_dupd(int);
static int  is_serial(int);
static void sendbreak(void);
static void senq(int);
static int  serial_read1(int, char *, int);
static int  set_mode( unsigned long, unsigned);
static void sputch(int);
static void uinit_8259(unsigned int int_value);

/* 8250 register addresses */
#define LCNT_REG	(serial_iobase + 0)
#define HCNT_REG	(serial_iobase + 1)
#define IER_REG 	(serial_iobase + 1)
#define ISR_REG 	(serial_iobase + 2)
#define LCR_REG 	(serial_iobase + 3)
#define MCR_REG 	(serial_iobase + 4)
#define ST_REG		(serial_iobase + 5)

/* 8250 status register bits */
#define GETRDYBIT	0x01
#define PUTRDYBIT	0x20

/* 8250 interrupt register bits */
#define ENTXINT 	0x02
#define ENRXINT 	0x01

/* 8250 line control register bits */
#define ENCNTREG	0x80
#define SETBREAK	0x40
#define CHLEN1		0x02
#define CHLEN0		0x01

/* 8250 modem control register btis */
#define OUT2		0x08
#define RTS		0x02		/* inverted */	
#define DTR		0x01		/* inverted */


/*
 * TIMER ROUTINES
 */
#define TKPSEC		17
#define MSECPTK 	(1000/TKPSEC)
#define TMRVEC		0x1c

static int bad; 	/* bad pointer, in case of bad timeout pointer */
static int *timeoutp = &bad;		/* timeout flag pointer */
static unsigned short timecnt = 0;	/* timeout ticks remaining */
static int inited = 0;
static void (*prev_timer_int)() = 0;
static int occurcnt = 0;		/* debug, time out count */

/******************************************************************************
 * settimeout()
 *	Set a "background" timer which sets requested flag when alloted 
 *	interval expires.
 *
 *	WARNING: Normal usage would settimeout( > 0 ), process until some event 
 *	occurs OR this flag sets (indicating times up), then settimeout(0)
 *	in order to turn the timer off.
 *	This is VERY important, because if the requester flag is for instance, 
 *	a stack variable, and the caller RETURNs before the timeout; then, when 
 *	the timer does go off, it will set some, now defunct location.
 ******************************************************************************/
static void
settimeout(ms, flagp)
    unsigned ms;
    int *flagp;
{
	int pif;

	if (!inited) {
		init_timer();
	}
	pif = change_if(0);
	if (ms == 0) {
		timeoutp = &bad;
		timecnt = 0;
	} else {
		timeoutp = flagp;
		timecnt = (ms + (MSECPTK - 1)) / MSECPTK + 1;
	}
	change_if(pif);
}


/******************************************************************************
 * TICK()
 *	C interrupt service routine.
 ******************************************************************************/
#pragma interrupt(TICK)
static void
TICK(void)
{
	if (timecnt && !--timecnt) {
		*timeoutp = 1;
		++occurcnt;			/* for debug */
	}
	_chain_intr(prev_timer_int);
}


/******************************************************************************
 * init_timer()
 *	Initialize our timer support.  Return 1 on success, 0 on failure.
 ******************************************************************************/
static int
init_timer(void)
{
	if (inited) {
		return 0;
	}
	prev_timer_int = _dos_getvect(TMRVEC);
	_dos_setvect(TMRVEC, TICK);
	inited = 1;
	return 1;
}

/******************************************************************************
 * term_timer()
 *	Terminate our timer support, releasing any resources which were used.
 ******************************************************************************/
static int
term_timer(void)
{
	if (inited) {
		_dos_setvect(TMRVEC, prev_timer_int);
		inited = 0;
	}
}

static int
xltbaud(baud, baudmap)
    unsigned long baud;
    register struct baudmap *baudmap;
{
	for ( ; baudmap->baud; baudmap++){
		if (baudmap->baud == baud){
			return baudmap->code;
		}
	}
	return -1;
}

/******************************************************************************
 * sputch()
 *	Put a serial character, wait if not currently ready to send.
 ******************************************************************************/
static void
sputch(ch)
    int ch;
{

	while ( !(_inbyte(ST_REG) & PUTRDYBIT) ){
		;
	}
	_outbyte( serial_iobase, ch );
}


/******************************************************************************
 * sendbreak()
 *	Send a break "character" (RS232 low for more than a character period).
 ******************************************************************************/
static void
sendbreak()
{
	unsigned char	ch;
	int	cnt;

	ch = _inbyte( LCR_REG );
	_outbyte( LCR_REG, ch|SETBREAK );
	for ( cnt = BREAKMS*LOOPERMS; cnt; --cnt ){
		;
	}
	_outbyte( LCR_REG, ch );
}

/******************************************************************************
 * serial_open()
 *	Open the serial device passed in as a parameter and prepare it for IO.
 *	Return -1 on error (unable to initialize requested device).
 ******************************************************************************/
static int
serial_open()
{
	if (strncmp(serial_device, "com", 3) == 0) {
		int port = serial_device[3] - '0';
		if (serial_iobase == 0){
			serial_iobase = get_port(port);
		}
		if (serial_ivec == 0) {
			switch (port) {
			case 1: serial_ivec = IR4; break; /* COM1 interrupt */
			case 2: serial_ivec = IR3; break; /* COM2 interrupt */
			case 3: serial_ivec = IR5; break; /* COM3 interrupt */
			}
		}
	} else if (strcmp(serial_device, "aux") == 0) {
		if (serial_iobase == 0){
			serial_iobase = 0x3e8;
		}
		if (serial_ivec == 0){
			serial_ivec = IR5;
		}
	}

	if ( (serial_iobase == 0)
	||   (serial_ivec   == 0)
	||   !set_mode(serial_baud,serial_ivec)
	||   !init_timer()) {
		return -1;
	}
	return 0;
}

/******************************************************************************
 * serial_close()
 *	Close the serial device, terminating usage in current context.
 ******************************************************************************/
static void
serial_close()
{
	_outbyte( IER_REG, 0);		/* disable ints */
	_outbyte( MCR_REG, 0);
	uinit_8259(serial_ivec);
	term_timer();
}

static int
serial_read1(int port, char *bufptr, int want)
{
	int cnt;
	int pif;

	pif = change_if( 0 );
	if ((cnt = siq.q_in - siq.q_out) < 0){
		cnt = &siq.q_buf[QBUFSZ] - siq.q_out;	/* don't wrap */
	}
	change_if( pif );

	if (cnt == 0) {
		return 0;
	}
	if (want < cnt){
		cnt = want;
	}

	memcpy( bufptr, siq.q_out, cnt );

	pif = change_if( 0 );
	siq.q_cnt -= cnt;
	if ((siq.q_out += cnt) >= &siq.q_buf[QBUFSZ]) {
		siq.q_out = siq.q_buf;
	} else if (siq.q_in == siq.q_out) {
		siq.q_in = siq.q_out = siq.q_buf;	/* reduce wraps */
	}
	change_if( pif );
	return cnt;
}

/******************************************************************************
 * serial_read()
 *	Attempt to read N bytes from the serial device, waiting up to
 *	timo (time-out) milliseconds for data to become available.  Return:
 *		0-N	Actual number of bytes read, 0 if none.
 *		-1	error condition
 ******************************************************************************/
static int
serial_read(int port, char *buf, int size, unsigned timo)
{
	int actual;
	int flag;

	flag = 0;
	settimeout(timo, &flag);
	do {
		actual = serial_read1(port, (char *)buf, size);
	} while ((actual == 0) && !flag);
	settimeout(0, &flag);
	return actual;
}

/******************************************************************************
 * serial_write()
 *	Attempt to write N bytes from the serial device, stopping as soon 
 *	as unable to send more without waiting.  Return:
 *		0-N	Actual number of bytes written, 0 if none.
 *		-1	error condition
 ******************************************************************************/
static int
serial_write( port, bufptr, want )
    int   port;
    char *bufptr;
    int   want;
{
	int i;

	for (i = 0; i < want; i++){
		sputch(*bufptr++);
	}
	return want;
}



/******************************************************************************
 * set_mode()
 *	Set the serial device to a specified operating mode.
 *	Return 0 on success, -1 on failure
 ******************************************************************************/
static int
set_mode(baud,ivec)
    unsigned long baud; /* Buad rate	*/
    unsigned ivec;	/* Interrupt vector */
{
	int val;

	if ((val = xltbaud(baud, baudmap)) <= 0){
		return 0;
	}
	_outbyte( LCR_REG, ENCNTREG );		/* enable count regs */
	_outbyte( LCNT_REG, (unsigned char)val);
	_outbyte( HCNT_REG, (unsigned char)(val >>8) );
	_outbyte( LCR_REG, CHLEN1 | CHLEN0 );	/* n,8,1 */
	_outbyte( MCR_REG, OUT2 | RTS | DTR );

	if (baud >= 57600) {
		/* Enable receive and transmit FIFOs.
		 *
		 * FCR<7:6>	00	trigger level = 1 byte
		 * FCR<5:4>	00	reserved
		 * FCR<3>	0	mode 0 - interrupt on data ready
		 * FCR<2>	0
		 * FCR<1>	0
		 * FCR<0>	1	turn on fifo mode
		 */
		_outbyte(ISR_REG, 0x01);
	}

	/* initialize serial queue and chip */
	siq.q_cnt = siq.q_ovfl = 0;
	siq.q_in = siq.q_out = siq.q_buf;
	init_8259(ivec);
	_outbyte(IER_REG, ENRXINT); /* enable ints */
	_outbyte(IER_REG, ENRXINT); /* do twice because of chip bug */
	return 1;
}


/******************************************************************************
 * S_INT()
 *	C serial interrupt routine (can't use debug).
 *	Assumes that only Data ready interrupt is enabled.
 ******************************************************************************/
#pragma interrupt(S_INT)
void
S_INT()
{
	int c;
	register struct queue *qp = &siq;
	int pif;
	
	pif = change_if(0);
	while (_inbyte(ST_REG) & GETRDYBIT) {
		c = _inbyte(serial_iobase);
		if (qp->q_cnt >= (QBUFSZ-1)) {
			qp->q_ovfl++;
			return;
		}
		*qp->q_in++ = c;
		qp->q_cnt++;
		if (qp->q_in >= &qp->q_buf[QBUFSZ]){
			qp->q_in = qp->q_buf;	/* pre-advance pointer */
		}
	}
	_outbyte(0x20, 0x20);
	change_if(pif);
}


/******************************************************************************
 * init_8259()
 *	Initailize the PCDOS 8259 serial vector for interrupt driven usage.
 *	return Boolean indicating success.
 ******************************************************************************/
static void
init_8259(ivec)
    unsigned ivec;	/* Interrupt vector */
{
	int pif;

	if (!orig_serial_vec) {
		orig_serial_vec = _dos_getvect(ivec);
		_dos_setvect(ivec, S_INT);
	}
	pif = change_if( 0 );
	_outbyte( IMR_8259, (_inbyte(IMR_8259) & ~(1 << ivec - 8)) );
	change_if( pif );
}


/******************************************************************************
 * uinit_8259()
 *	Unset the PCDOS 8259 interrupt controller serial interrupt vector 
 *	used for interrupt mode.  Return Boolean indicating success.
 ******************************************************************************/
static void
uinit_8259(ivec)
    unsigned ivec;	/* Interrupt vector */
{
	int pif;

	pif = change_if( 0 );
	_outbyte( IMR_8259, (_inbyte(IMR_8259) | (1 << ivec - 8)) ); /* mask */
	change_if( pif );
	if (orig_serial_vec) {
		_dos_setvect(ivec, orig_serial_vec);
		orig_serial_vec = 0;
	}
}

/******************************************************************************
 * get_port()
 *	equipment check bios call (bit #)
 *		15-14	printers
 *		   12	game io
 *		 11-9	com devices
 *		  7-6	disk drives (0=1)
 *		  5-4	init video mode
 *		  3-2	ram size
 *		    0	are any disk drives
 ******************************************************************************/
static int
get_port(comwhich)
    int comwhich;
{
	union REGS regs;
	int combase;

	int86( 0x11, &regs, &regs );
	if (comwhich > ((regs.x.ax >> 9) & 7)){
	    return 0;
	}
	combase = ((unsigned short *)0x400)[comwhich-1];
	return combase;
}


int
change_if(int f)
{
	volatile unsigned int old_f = _getflags();
	_setflags(f ? old_f | _FLAG_INTERRUPT : old_f & ~_FLAG_INTERRUPT);
	return (old_f & _FLAG_INTERRUPT) != 0;
}





