/****************** QMEM - Memory management routines **************
* Copyright 1986 Eric Balkan                                       *
* Packet Press, 14704 Seneca Castle Ct., Gaithersburg, MD 20878    *
*                                                                  *
* Permission is given to incorporate these functions into a larger *
* work and to distribute the resulting object code or executable   *
* module. The author reserves all other rights, including but not  *
* limited to, the exclusive right to distribute the source code.   *
*                                                                  *
* Updated & Modified 8/88 by C Gazette Staff                       *
* Compilers: Microsoft C 5.1 - compile as is                       *
*            Turbo C 1.5 - compile as is, but be sure to compile   *
*               with the -K option so that chars are unsigned.     *
*               TC cannot handle the '#define char unsigned char'  *
*               that is used in the Microsoft version              *
*            DeSmet C 3.1   - compile as is                        *
*                                                                  *
* Memory Models: Supports both 16- and 32-bit pointers. Define the *
*             preprocessor variable LARGE_PTR for 32-bit pointers. *
*                                                                  *
* Preprocessor switches:                                           *
*     DIAGNOSTICS - includes memdiag()                             *
*     DRIVER - includes main() for testing                         *
*     LARGE_PTR - 32-bit pointers                                  *
*                                                                  *
*******************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define DRIVER

#if defined DRIVER
#define DIAGNOSTICS
#endif

#if !defined(__TURBOC__)        /* Force all cahrs to be unsigned, so */
#define char unsigned char      /* they can be used a s subscripts.   */
#endif

/* select memory model */

/* Turbo C */
#if defined __TURBOC__
    #if defined(__COMPACT__) || defined(__LARGE__) || defined(__HUGE__)
    #define LARGE_PTR
    #endif
#endif

/* Microsoft C */
#if defined(M_I86CM) || defined(M_I86LM) || defined(M_I86HM)
#define LARGE_PTR
#endif

/* DeSmet C */
#if defined(__LARGE__)
#define LARGE_PTR
#endif

#ifndef min
#define min(x,y) ((x) < (y) ? (x) : (y))
#endif

#define MAXARRAY 100

struct MEMTAB {
	char     *loc;           /* location of array (first element) */
	int      size;           /* size of each element */
	unsigned totsize;        /* total size of array */
	int      num;            /* number of elements in array */
	int      num_inuse;      /* number of elements in use */
	int      high_water;     /* number of elements used at any one time */
	int      ext_no;         /* table number of extension */
	int      ext_amt;        /* number of elements to get as an extension */
	int      flags;          /* & IS_EXT ==> this is an extension */
	char     *bitmap;        /* bitmap of allocated elements */
	unsigned mapsize;        /* size of bitmap */
};

#define IS_EXT 1             /* status bit for flags in struct MEMTAB */

static struct MEMTAB ctl[MAXARRAY];  /* one MEMTAB per array, up to some max */
static int next_memtab = 0;          /* next array to use */
static char bit_value[] = {0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01,0x00};
static char bit_byte[]  = {0x00,0x80,0xC0,0xE0,0xF0,0xF8,0xFC,0xFE};

void *qmalloc(), *qcalloc(), *qrealloc(), *onbit(), qmemstop();
int qmemstart();
void *qfree();
int gettyps(), gettypa(), offbit(), testbit();

/*
 *   QMEMSTART - allocate an array
 *   Returns an array number or an error
 *    -1  maximum number of arrays in use
 *    -2  no memory available for a bitmap
 *    -3  no memory available for another array
 *    -4  bad parameters
 */

int qmemstart (size, num, ext_amt)
int size;                    /* Size of each member */
int num;                     /* Number of members */
int ext_amt;                 /* Number of elements to put in extension */
{
	char *loc;
	char *bitmap;
	unsigned totsize;
	unsigned mapsize;
	int free_slot;
	int rmdr;

	/* find a spot */
	for (free_slot = 0; free_slot < next_memtab; free_slot++)
		if (!ctl[free_slot].num)
			break;

	if (free_slot == next_memtab)
		if (next_memtab > MAXARRAY)
			return (-1);

	if (!size || !num)
		return (-4);
	totsize = size * num;
	if ((loc = (char *) calloc (1, totsize)) != NULL) {
		mapsize = (num >> 3) + 1;
		if ((bitmap = (char *) malloc (mapsize)) == NULL)
			return (-2);
		memset (bitmap, 0xff, mapsize);
		/* turn off extra ones at end */
		rmdr = num % 8;
		bitmap [mapsize - 1] = bit_byte [rmdr];

		ctl[free_slot].loc        = loc;
		ctl[free_slot].size       = size;
		ctl[free_slot].totsize    = totsize;
		ctl[free_slot].num        = num;
		ctl[free_slot].num_inuse  = 0;
		ctl[free_slot].high_water = 0;
		ctl[free_slot].ext_no     = -1;
		ctl[free_slot].ext_amt    = ext_amt;
		ctl[free_slot].flags      = 0;
		ctl[free_slot].bitmap     = bitmap;
		ctl[free_slot].mapsize    = mapsize;
		if (free_slot == next_memtab)
			next_memtab++;
		return (free_slot);
	}
	return (-3);
}

/*
 * QMEMSTOP - free memory associated with an array
 */
void qmemstop (curtab)
int curtab;
{
	if (ctl[curtab].ext_no != -1)           /* free any associated arrays */
		qmemstop (ctl[curtab].ext_no);
	if (ctl[curtab].bitmap != NULL)
		free ((void *) (ctl[curtab].bitmap));
	if (ctl[curtab].loc != NULL)
		free ((void *) (ctl[curtab].loc));
	ctl[curtab].bitmap = NULL;
	ctl[curtab].loc = NULL;
	ctl[curtab].num = 0;
	ctl[curtab].ext_no = -1;
}

/*
 * QMALLOC - allocates memory
 * Returns address of memory allocated or error
 *   Null - no memory available
 */
void *qmalloc (size)
unsigned size;
{
	void *ptr;
	int curtab;

	if ((curtab = gettyps(size)) == -1) /* doesn't fit */
		ptr = malloc (size);
	else if ((ptr = onbit (curtab)) == NULL)   /* find free slot in bitmap */
		puts("qmalloc: count/usage mismatch\n");
	return (ptr);
}

/*
 * GETTYPS - given a size, returns an array number
 *
 *	Technique:  we search the list of memory arrays looking for
 *  an acceptable fit.  There are three possible
 *  results of this search:
 *
 *  1) acceptable fit found and it has room for us
 *  2) acceptable fit found, but it is full
 *  3) no acceptable fit found
 *
 *  For each possiblity, we return the following:
 *
 *  1) array number of acceptable fit
 *  2) we attempt to allocate a new array.  if successful, return
 *     the number of the new array.  else, return -1
 *  3) -1
 *
 *  An acceptable fit is currently defined as a match that wastes
 *  less than 100 bytes or one that wastes less than 25% of the
 *  memory element.
 */
int gettyps (size)
unsigned size;
{
	int curtab;
	int i, error;
	int ok_fit, ok_off, best_fit, best_off;

	curtab = -1;
	ok_fit = -1;	/* tracks acceptable fits with room */
	best_fit = -1;	/* tracks best acceptable fit, even if no room */

	for (i=0; i<next_memtab; i++)
		if (ctl[i].num && ctl[i].size >= size) { /* could be ok */
			error = ctl[i].size - size;
			if (error < (ctl[i].size >> 2) || error < 100) {
				if (ctl[i].num_inuse < ctl[i].num) /* acceptable, with room */
					if (ok_fit == -1 || error <= ok_off) {
						ok_fit = i;
						ok_off = error;
					}
				if (best_fit == -1 || error <= best_off) {
					best_fit = i;
					best_off = error;
				}
			}
		}

	/* now, return ok_fit or else allocate space at best_fit */
	if (ok_fit != -1)
		curtab = ok_fit;
	else if (best_fit != -1 && ctl[best_fit].ext_amt) { /* make new entry */
		i = qmemstart(ctl[best_fit].size, ctl[best_fit].num,
			ctl[best_fit].ext_amt);
		if (i >= 0) {
			curtab = i;
			ctl[i].flags |= IS_EXT;
			ctl[best_fit].ext_no = i;
		}
	}
    return (curtab);
}


/*
 * GETTYPA - given an address, returns an
 *        array number
 */
int gettypa (ptr)
void *ptr;
{
	int curtab = -1;
	int i;
	#if defined(LARGE_PTR)
	unsigned ptr_to_seg(void *);
	#endif

	for (i = 0; i < next_memtab; ++i) {
		if (ctl[i].num)
			#if defined (LARGE_PTR)
			if (ptr_to_seg((void *) ctl[i].loc) == ptr_to_seg((void *) ptr))
			#endif
			if (ctl[i].loc == ptr || (ctl[i].loc < ptr
				&& (ctl[i].loc + ctl[i].totsize > ptr))) {
				curtab = i;
				break;
			}
	}
	return (curtab);
}

#if defined (LARGE_PTR)
/* pointer manipulation routines - we use these to avoid weaknesses in
   FP_SEG and FP_OFF macros supplied with some compilers */

typedef union {		/* variants on 32 bits */
	void *ptr;	/* a pointer */
	struct {	/* or as two unsigned words: */
		unsigned offset;	/* an offset ... */
		unsigned segment;	/* ... and its segment */
	} split;
} bits_32;

unsigned ptr_to_offset(x)	/* return offset portion of a pointer */
void *x;
{
	bits_32 y;

	y.ptr = x;
	return(y.split.offset);
}

unsigned ptr_to_seg(x)	/* return the segment portion of a pointer */
void *x;
{
	bits_32 y;

	y.ptr = x;
	return(y.split.segment);
}
#endif

/*
 *  QCALLOC - allocates and clears memory
 *  Returns address of memory allocated or error
 *    Null - no memory available
 */
void *qcalloc (num, size)
int num;
int size;
{
	char *ptr;
	unsigned amt = size * num;

	if ((ptr = qmalloc (amt)) != NULL)
		memset (ptr, 0, amt);
	return (ptr);
}

/*
 * QREALLOC - re-allocates memory
 * Returns address of memory allocated or error
 *   Null - no memory available
 */
void *qrealloc (op, size)
void *op;                /* old pointer */
unsigned size;                /* number of bytes to reallocate */
{
	void *np;             /* new pointer */
	int curtab;

	if (op == NULL)
		return(qmalloc(size));

    if ((curtab = gettypa(op)) != -1) { /* it's one of our blocks */
		if ((np = qmalloc(size)) != NULL) {
			memmove (np, op, (unsigned) min(ctl[curtab].size, size));
			qfree (op);
		}
	}
	else /* allocated by DOS - let it figure it out */
		np = realloc(op, size);

	return(np);
}

/*
 * QFREE - frees an element
 * Returns address of memory freed
 */
void *qfree (ptr)
void *ptr;
{
	int curtab;

	if (ptr != NULL) {
		if ((curtab = gettypa (ptr)) == -1)
			free (ptr);
		else if (testbit (ptr, curtab) == 0) /* 0 if allocated */
			offbit (ptr, curtab);
	}
	return (ptr);
}

/*
 *  TESTBIT - test whether an element is in use
 *  Returns:
 *     0 if in use
 *    -1 if not in use
 */
int testbit (ptr, curtab)
int curtab;
void *ptr;
{
	unsigned handle;
	int which_byte;
	int which_bit;
	char pos;

	if (! ctl[curtab].size)
		return (-1);
#if defined(LARGE_PTR)
	/* we specially manipulate the pointer offset to avoid sign errors when
	   dealing with objects large than 32K */
	handle = ptr_to_offset((void *) ptr) -
		ptr_to_offset((void *) ctl[curtab].loc);
#else
	handle = ((char *) ptr) - ctl[curtab].loc;
#endif

	handle = handle / ctl[curtab].size;
	which_byte = handle >> 3;
	which_bit  = handle % 8;
	pos = ctl[curtab].bitmap[which_byte];
	if (bit_value[which_bit] & pos)
		return (-1);
	return (0);                       /* allocated */
}

/*
 *  ONBIT - finds an element to use
 *  Returns: - pointer to element
 *           - NULL if no elements available
 */
void *onbit (curtab)
int curtab;
{                     /* 1 = element free */
	char *bitmap;
	int i, j, k;
	unsigned pattern;

	bitmap = ctl[curtab].bitmap;
	for (i = 0; i < ctl[curtab].mapsize; ++i) {
		if (!bitmap[i])           /* these 8 taken */
			continue;

		/* there is at least one free bit - now find it */
		j = i << 3;
		pattern = 0x80;
		for (k=0; k<8; k++)
			if (bitmap[i] & pattern) {
				bitmap[i] &= ~pattern;
				j += k;
				break;
			}
			else
				pattern = pattern >> 1; /* try next bit */

		if (ctl[curtab].high_water < ++ctl[curtab].num_inuse)
			ctl[curtab].high_water = ctl[curtab].num_inuse;
		return((j * ctl[curtab].size) + ctl[curtab].loc);
	}
	return (NULL);              /* none free */
}

/*
 *  OFFBIT - finds the element to free
 *  Returns:
 *  pointer to element
 *  -1 on error
 */
int offbit (ptr, curtab)
char *ptr;
int curtab;
{
	char *bitmap;
	int which_byte;
	int which_bit;
	int handle;

	if (!ctl[curtab].size)
		return (-1);
	handle = (ptr - ctl[curtab].loc) / ctl[curtab].size;
	bitmap = ctl[curtab].bitmap;
	which_byte = handle >> 3;
	which_bit  = handle % 8;
	--ctl[curtab].num_inuse;
	bitmap[which_byte] |= bit_value[which_bit];
	return (0);
}

#if defined DIAGNOSTICS
/*
 *  MEMDIAG - statistics/diagnostics display
 */
void memdiag()
{
	int i, j, k;

	puts ("\n");
	for (i = 0; i < next_memtab; ++i) {
		puts ("#   loc     size   num   tsize  in-use hwmk ");
		puts ("flgs extno extamt mapsize bitmap\n");
		if (!ctl[i].num)
			continue;
		#if defined(LARGE_PTR)
		printf ("%d %p ", i, ctl[i].loc);
		#else
		printf ("%d %4x  ", i, ctl[i].loc);
		#endif
		printf ("%5d", ctl[i].size);
		printf (" %5d", ctl[i].num);
		printf ("   %5d", ctl[i].totsize);
		printf ("   %5d", ctl[i].num_inuse);
		printf ("%5d", ctl[i].high_water);
		printf ("    %01x", ctl[i].flags);
		printf ("   %3d", ctl[i].ext_no);
		printf ("  %5d", ctl[i].ext_amt);
		printf ("    %4d ",ctl[i].mapsize);
		for (j = 0; j < ctl[i].mapsize; ++j)
			printf ("%.2x", ctl[i].bitmap[j]);
		for (j = 0; j < ctl[i].high_water; ++j) {
			if (testbit (ctl[i].loc + (j * ctl[i].size), i))
				continue;                           /*not allocated */
			#if defined(LARGE_PTR)
			printf ("\n %p: ", ctl[i].loc + (j * ctl[i].size));
			#else
			printf ("\n %2x: ", ctl[i].loc + (j * ctl[i].size));
			#endif
			for (k = 0; k < ctl[i].size && k < 16; ++k) {
				if (k % 4 == 0)
					putchar (' ');
				printf ("%02x", *(ctl[i].loc + (j * ctl[i].size) + k));
			}
			putchar (' ');
			putchar ('|');
			for (k = 0; k < ctl[i].size && k < 24; ++k) {
				if (*(ctl[i].loc + (j * ctl[i].size) + k) > 0x1f)
					putchar (*(ctl[i].loc + (j * ctl[i].size + k)));
				else
					putchar ('.');
			}
			putchar ('|');
		}
		puts ("\n");
	}
}

#endif

#if defined DRIVER

void main()
{
	char inline[100];
	int size, num, ext_amt;
	void *ptr, *p2;

	for(;;) {
        puts
        ("(I)nit or (D)elete array, (A)llocate/(F)ree/(R)ealloc, (S)tatus ?");
		gets(inline);
		if (strlen(inline) == 0)
			break;

		switch(tolower(inline[0])) {
			case 'i':
				puts("size num ext_amt: ");
				gets(inline);
				if (sscanf(inline, "%d %d %d", &size, &num, &ext_amt) == 3)
                    printf("qmemstart() returned %d\n",
                     qmemstart(size, num, ext_amt));
				break;

			case 'd':
				puts("table#: ");
				gets(inline);
				if (sscanf(inline, "%d", &num) == 1) {
					qmemstop(num);
					printf("qmemstop() invoked\n");
				}
				break;

			case 'a':
				puts("size: ");
				gets(inline);
				ptr = NULL;
				if (sscanf(inline, "%d", &size) == 1)
					#if defined LARGE_PTR
                    printf("qmalloc() returned %p\n",
                         ptr=qmalloc((unsigned) size));
					#else
                    printf("qmalloc() returned %X\n",
                         ptr=qmalloc((unsigned) size));
					#endif
				if (ptr != NULL)
					strcpy(ptr, inline);
				break;

			case 'r':
				puts("pointer size: ");
				gets(inline);
				p2 = NULL;
				#if defined LARGE_PTR
				if (sscanf(inline, "%p %d", &ptr, &size) == 4)
                    printf("qrealloc() returned %p\n",
                        p2=qrealloc(ptr, (unsigned) size));
				#else
				if (sscanf(inline, "%X %d", &ptr, &size) == 2)
                    printf("qrealloc() returned %X\n",
                        p2=qrealloc(ptr, (unsigned) size));
				#endif
				if (p2 != NULL)
					strcpy(p2, inline);
				break;

			case 'f':
				puts("pointer: ");
				gets(inline);
				#if defined LARGE_PTR
				if (sscanf(inline, "%p", &ptr) == 3)
					printf("qfree() returned %p\n", qfree(ptr));
				#else
				if (sscanf(inline, "%X", &ptr) == 1)
					printf("qfree() returned %X\n", qfree(ptr));
					#endif
				break;

			case 's':
				memdiag();
				break;
		}
	}
}
#endif

