/***************************** SCLFONT.C *****************************
*                       L I S T I N G   T W O                        *
* DESCRIPTION:  Program to resize GEM font files                     *
* COMPILER(s):  Borland 2.0+, Lattice 3.4+, MS 5.0+, Zortech C/C++   *
* AUTHOR:       Marv Luse, Ithaca Street Software, Inc.              *
* DATE:         June, 1989                                           *
* COPYRIGHT:    Use freely as long as authorship is acknowledged     *
*********************************************************************/

#include <stdlib.h>
#include <stdio.h>

/* functions that might be in a compiler-dependent header file */

void * memset( void *, int, unsigned int );
int    getch( void );
char * strcpy( char *, char * );

/* definition for GEM font header - see Listing 1 */
#include "gem.h"

/* macro to test a bit */
#define BitIsOn( c, b )   ( (c) & (0x80 >> (b)) )

/* macro to set a bit */
#define SetBit( c, b )   ( (c) | (0x80 >> (b)) )

/* outrageous value */
#define UNINITIALIZED   -12345

/* limits on new character cell size - values are   */
/* not critical - used to identify erroneous values */
#define UPPER_LIMIT   300
#define LOWER_LIMIT   5

/*--------------------------------------------------------------------*/
/* Globally referenced variables...                                   */

gem_header   gm_hdr;                     /* font file header */
int         *old_cot, *new_cot,          /* char offset tables */
             old_wid, new_wid,           /* char cell pix widths */
             old_hgt, new_hgt,           /* char cell pix heights */
             old_end, new_end,           /* bitmap pix widths */
             char_cnt;                   /* number of chars in font */
char       **old_bmap, **new_bmap,       /* scan line pointer lists */
            *gm_tbls,                    /* area from hdr to bmap */
             in_font[80],                /* input file name */
             out_font[80];               /* output file name */

/*--------------------------------------------------------------------*/
/* Display a message and exit...                                      */

void exit_pgm( char *msg, int retv )
{
    printf( "\n%s", msg );
    exit( retv );
}

/*--------------------------------------------------------------------*/
/* Explain program usage...                                           */

void explain_pgm( void )
{
    printf( "\n" );
    printf( "\n---------- GEM Font File Scaling Program ----------" );
    printf( "\n" );
    printf( "\nusage:" );
    printf( "\n" );
    printf( "\n    SCLFONT  in_font  [out_font]  [w=nnn]  [h=nnn]" );
    printf( "\n" );
    printf( "\nwhere:" );
    printf( "\n" );
    printf( "\n    in_font   ::  path to input font file" );
    printf( "\n                  (must be specified)" );
    printf( "\n    out_font  ::  path to output font file" );
    printf( "\n                  (defaults to in_font)" );
    printf( "\n    w=nnn     ::  set new char width to nnn pixels" );
    printf( "\n                  (defaults to old width)" );
    printf( "\n    h=nnn     ::  set new char height to nnn pixels" );
    printf( "\n                  (defaults to old height)" );
    printf( "\n" );
    printf( "\n---------------------------------------------------" );
}

/*--------------------------------------------------------------------*/
/* Initialize values from the command line...                         */

void initialize( int argc, char *argv[] )
{
    int  i;
    char c1, c2;

    /* set defaults... */
    in_font[0] = 0;
    out_font[0] = 0;
    new_wid = new_hgt = UNINITIALIZED;

    for( i = 1; i < argc; i++ )
    {
        c1 = *argv[i];
        c2 = *(argv[i] + 1);
        if ( ((c1=='w') || (c1=='W'))  && (c2=='=') )
            sscanf( argv[i]+2, "%d", &new_wid );
        else if ( ((c1=='h') || (c1=='H'))  && (c2=='=') )
            sscanf( argv[i]+2, "%d", &new_hgt );
        else if ( in_font[0] == 0 )
            strcpy( in_font, argv[i] );
        else if ( out_font[0] == 0 )
            strcpy( out_font, argv[i] );
        else
            exit_pgm( "Invalid token on command line", 8 );
    }
}

/*--------------------------------------------------------------------*/
/* Check for the existence of the specified file (size > 0)...        */

int  file_exists( char * path )
{
    int  c;
    FILE *fp;
    fp = fopen( path, "rb" );
    if ( fp == NULL ) return( 0 );
    c  = fgetc( fp );
    fclose( fp );
    return( (c == EOF) ? 0 : 1 );
}

/*--------------------------------------------------------------------*/
/* Verify values from command line...                                 */

void verify_args( void )
{
    int i;
    /* check file(s) involved */
    if ( in_font[0] == 0 )
        exit_pgm( "Input font file not specified.", 8 );
    if ( ! file_exists( in_font ) )
        exit_pgm( "Input font file not found.", 8 );
    if ( out_font[0] == 0 )
        strcpy( out_font, in_font );
    /* okay to overwrite an existing file ? */
    if ( file_exists( out_font ) )
    {
        printf( "\nOkay to overwrite %s (Y|N)?", out_font );
        while ( 1 )
        {
            i = getch();
            if ( (i=='y') || (i=='Y') )
                break;
            if ( (i=='n') || (i=='N') )
                exit_pgm( "Execution cancelled.", 8 );
        }
    }
    /* check new width and height, if specified */
    if ( new_wid != UNINITIALIZED )
        if ( (new_wid < LOWER_LIMIT) || (new_wid > UPPER_LIMIT) )
            exit_pgm( "New width is invalid.", 8 );
    if ( new_hgt != UNINITIALIZED )
        if ( (new_hgt < LOWER_LIMIT) || (new_hgt > UPPER_LIMIT) )
            exit_pgm( "New height is invalid.", 8 );

    /* at least one or the other should have been specified */
    if ( (new_wid == UNINITIALIZED) && (new_hgt == UNINITIALIZED) )
        exit_pgm( "New width and/or height must be specified.", 8 );
}

/*--------------------------------------------------------------------*/
/* Scale an integer value without overflow...                         */

int scl_int( int i, int m, int d )
{
    long v;
    v = i;  v *= m;  v += d/2;  v /= d;
    return( (int) v );
}

/*--------------------------------------------------------------------*/
/* Check specified pixel range in source bitmap for any bit set...    */

int any_pixel_set( int toprow, int botrow, int leftcol, int rightcol )
{
    char *row;
    int   i, j, bit;
    for( i = toprow; i <= botrow; i++ )
    {
        row  = *(old_bmap + i);
        row += leftcol / 8;
        bit  = leftcol % 8;
        for( j = leftcol; j <= rightcol; j++ )
        {
            if ( BitIsOn( *row, bit ) ) return( 1 );
            bit++;
            if ( bit > 7 )
            {
                bit=0; row++;
            }
        }
    }
    return( 0 );
}

/*--------------------------------------------------------------------*/
/* Fill in the new font's bitmap (should already be zero-filled)...   */

void fill_new_bmap( void )
{
    int   i, j, x1, x2, y1, y2, bit, row_inc;
    char *row;

    /* compute how often to write a dot to stdout to  */
    /* indicate that something is indeed happening    */
    row_inc = new_hgt / 8 + 1;
    printf( "\nScaling" );

    /* loop for each scan line (row) */
    for( i=0; i<new_hgt; i++ )
    {
        row = *(new_bmap + i);
        bit = 0;

        /* determine this pixel's row range on source bitmap */
        y1 = scl_int( i, old_hgt, new_hgt );
        y2 = scl_int( i+1, old_hgt, new_hgt );

        /* examine at least 1 row */
        if ( y2 > y1 )  y2--;

        /* loop for each column in current row */
        for( j = 0; j < new_cot[char_cnt]; j++ )
        {
            /* determine this pixel's col range on source bitmap */
            x1 = scl_int( j, old_end, new_end );
            x2 = scl_int( j+1, old_end, new_end );
            /* examine at least 1 col */
            if ( x2 > x1 )  x2--;
            /* do we set this pixel ? */
            if ( any_pixel_set( y1, y2, x1, x2 ) )
                *row = SetBit( *row, bit );
            /* next bit... */
            bit++;
            if ( bit > 7 )
            {
                bit=0; row++;
            }
        }
        if ( (i+1) % row_inc == 0 )
            printf( "." );
    }
    printf( "done" );
}

/*--------------------------------------------------------------------*/
/* Read the old font file...                                          */

void read_old_font( char *path )
{
    int   i, nbytes;
    FILE *inp;
    printf( "\nReading '%s'...", path );
    inp = fopen( path, "rb" );
    if ( inp == NULL )
        exit_pgm( "Error opening input font file.", 8 );

    /* read the header and set up for rest of file... */
    if ( fread( &gm_hdr, sizeof( gem_header ), 1, inp ) != 1 )
        exit_pgm( "Error reading font header.", 8 );

    /* read portion of file between header and bitmap... */
    nbytes = (int) ( gm_hdr.map_ofs - sizeof( gem_header ) );
    gm_tbls = (char *) malloc( nbytes );
    if ( gm_tbls == NULL )
        exit_pgm( "Error - insufficient memory.", 8 );
    if ( fread( gm_tbls, nbytes, 1, inp ) != 1 )
        exit_pgm( "Error reading font tables.", 8 );

    /* allocate and read the char bitmap's scan lines... */
    old_bmap = (char **) malloc( gm_hdr.form_height * sizeof( char * ) );
    if ( old_bmap == NULL )
        exit_pgm( "Error - insufficient memory.", 8 );
    for( i = 0; i < gm_hdr.form_height; i++ )
    {
        old_bmap[i] = (char *) malloc( gm_hdr.form_width );
        if ( old_bmap[i] == NULL )
            exit_pgm( "Error - insufficient memory.", 8 );
        if ( fread( old_bmap[i], gm_hdr.form_width, 1, inp ) != 1 )
            exit_pgm( "Error reading font bitmap.", 8 );
    }
    fclose( inp );
    printf( "done" );
}

/*--------------------------------------------------------------------*/
/* Initialize font variables, allocate new bitmap...                  */

void init_font_vars( void )
{
    int  i, n, nbytes;
    old_wid  = gm_hdr.cell_width;
    old_hgt  = gm_hdr.form_height;

    /* check for default values */
    if ( new_wid == UNINITIALIZED ) new_wid = old_wid;
    if ( new_hgt == UNINITIALIZED ) new_hgt = old_hgt;
    old_cot  = (int *)( (char *)(gm_tbls + gm_hdr.cot_ofs -
        sizeof( gem_header )) );
    char_cnt = gm_hdr.max_char - gm_hdr.min_char + 1;
    new_cot  = (int *) malloc( sizeof( int ) * (char_cnt + 1) );
    if ( new_cot == NULL )
        exit_pgm( "Error - insufficient memory.", 8 );

    /* set bitmap widths (in pixels) */
    old_end = old_cot[char_cnt];
    new_end = scl_int( old_end, new_wid, old_wid );

    /* fill in new c.o.t. */
    for ( i=0; i <= char_cnt; i++ )
    {
        new_cot[i] = scl_int( old_cot[i], new_end, old_end );
        /* check that rounding doesn't clip first column */
        n = scl_int( new_cot[i], old_end, new_end );
        if ( n > old_cot[i] ) new_cot[i]--;
    }

    /* allocate space for new bitmap... */
    new_bmap = (char **) malloc( new_hgt * sizeof( char * ) );
    if ( new_bmap == NULL )
        exit_pgm( "Error - insufficient memory.", 8 );

    /* scan line length */
    nbytes = (new_end - 1) / 8 + 1;
    for( i=0; i<new_hgt; i++ )
    {
        new_bmap[i] = (char *) malloc( nbytes );
        if ( new_bmap[i] == NULL )
            exit_pgm( "Error - insufficient memory.", 8 );
        memset( new_bmap[i], 0, nbytes );
    }
}

/*--------------------------------------------------------------------*/
/* Write the new font file...                                         */

void write_new_font( char *path )
{
    int   i, nbytes;
    FILE *outp;

    /* update the header to reflect changes */
    gm_hdr.cell_width  = new_wid;
    gm_hdr.char_width  = scl_int( gm_hdr.char_width, new_wid, old_wid );
    gm_hdr.form_height = new_hgt;
    gm_hdr.ascent      = scl_int( gm_hdr.ascent, new_hgt, old_hgt );
    gm_hdr.descent     = new_hgt - gm_hdr.ascent;
    gm_hdr.form_width  = ( new_cot[char_cnt] - 1 ) / 8 + 1;

    /* update the old c.o.t. from the new */
    for( i=0; i<=char_cnt; i++ )
        old_cot[i] = new_cot[i];

    /* write the file */
    printf( "\nWriting '%s'...", path );
    outp = fopen( path, "wb" );
    if ( outp == NULL )
        exit_pgm( "Error opening output font file.", 8 );
    if ( fwrite( &gm_hdr, sizeof( gem_header ), 1, outp ) != 1 )
        exit_pgm( "Error writing font header.", 8 );
    nbytes = (int) ( gm_hdr.map_ofs - sizeof( gem_header ) );
    if ( fwrite( gm_tbls, nbytes, 1, outp ) != 1 )
        exit_pgm( "Error writing font tables.", 8 );
    nbytes = gm_hdr.form_width;
    for( i = 0; i <= new_hgt; i++ )
        if ( fwrite( new_bmap[i], nbytes, 1, outp ) != 1 )
            exit_pgm( "Error writing font bitmap.", 8 );
    fclose( outp );
    printf( "done" );
}

/*--------------------------------------------------------------------*/
/*                            M A I N                                 */

void main( int argc, char *argv[] )
{
    /* check for a help request */
    if ( (argc < 2) || (*argv[1] == '?') )
    {
        explain_pgm();
        exit_pgm( "", 0 );
    }

    /* process the request */
    initialize( argc, argv );
    verify_args();
    read_old_font( in_font );
    init_font_vars();
    fill_new_bmap();
    write_new_font( out_font );
    exit_pgm( "end of program", 0 );
}